/*
 * Copyright (c) Red Hat Inc.

 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Authors: Dave Airlie <airlied@redhat.com>
 */

/* simple list based uncached page allocator 
 * - Add chunks of 1MB to the allocator at a time.
 * - Use page->lru to keep a free list
 * - doesn't track currently in use pages
 * 
 *  TODO: Add shrinker support
 */

#include "drmP.h"
#include <asm/agp.h>

#include "drm_page_alloc.h"

static struct list_head uncached_free_list;
static struct list_head cached_free_list;

static struct mutex page_alloc_mutex;
static int page_alloc_inited;

struct drm_page_alloc_usage drm_page_alloc_data;

/* add 1MB at a time */
#define NUM_PAGES_TO_ADD 256

static inline void drm_page_put(struct page *page, int cached)
{
	if (!cached)
		unmap_page_from_agp(page);
	put_page(page);
	__free_page(page);
}

int drm_add_pages_locked(int num_pages, int cached)
{
	struct page *page;
	int i;

	DRM_DEBUG("adding %scached memory %ld\n", cached ? "" : "un",
		  num_pages * PAGE_SIZE);

	for (i = 0; i < num_pages; i++) {

		page = alloc_page(GFP_KERNEL | __GFP_ZERO | GFP_DMA32);
		if (!page) {
			DRM_ERROR("unable to get page %d\n", i);
			return i;
		}

		get_page(page);
		if (!cached) {
#ifdef CONFIG_X86
			set_memory_wc((unsigned long)page_address(page), 1);
#else
			map_page_into_agp(page);
#endif
			list_add(&page->lru, &uncached_free_list);
			drm_page_alloc_data.total_uncached_pages++;
			drm_page_alloc_data.uncached_pages_in_list++;
		} else {
			list_add(&page->lru, &cached_free_list);
			drm_page_alloc_data.total_cached_pages++;
			drm_page_alloc_data.cached_pages_in_list++;
		}
	}
	return i;
}

struct page *drm_get_page(int cached)
{
	struct page *page = NULL;
	int ret;
	struct list_head *free_list;

	free_list = cached ? &cached_free_list : &uncached_free_list;

	mutex_lock(&page_alloc_mutex);
	if (list_empty(free_list)) {
		ret = drm_add_pages_locked(NUM_PAGES_TO_ADD, cached);
		if (ret == 0)
			return NULL;
	}

	page = list_first_entry(free_list, struct page, lru);
	list_del(&page->lru);
	if (cached)
		drm_page_alloc_data.cached_pages_in_list--;
	else
		drm_page_alloc_data.uncached_pages_in_list--;

	mutex_unlock(&page_alloc_mutex);
	return page;
}

void drm_put_page(struct page *page, int cached)
{
	mutex_lock(&page_alloc_mutex);
	if (cached) {
		drm_page_alloc_data.cached_pages_in_list++;
		list_add(&page->lru, &cached_free_list);
	} else {
		drm_page_alloc_data.uncached_pages_in_list++;
		list_add(&page->lru, &uncached_free_list);
	}
	mutex_unlock(&page_alloc_mutex);
}

void drm_release_all_pages(int cached)
{
	struct page *page, *tmp;
	struct list_head *free_list;

	free_list = cached ? &cached_free_list : &uncached_free_list;

	list_for_each_entry_safe(page, tmp, free_list, lru) {
		list_del(&page->lru);
		drm_page_put(page, cached);
	}
	drm_page_alloc_data.total_cached_pages = 0;
	drm_page_alloc_data.total_uncached_pages = 0;
}

int drm_page_alloc_init(void)
{

	if (page_alloc_inited)
		return 0;

	INIT_LIST_HEAD(&uncached_free_list);
	INIT_LIST_HEAD(&cached_free_list);

	mutex_init(&page_alloc_mutex);
	page_alloc_inited = 1;
	return 0;

}

void drm_page_alloc_fini(void)
{
	if (!page_alloc_inited)
		return;
	
	page_alloc_inited = 0;
	drm_release_all_pages(0);
	drm_release_all_pages(1);
}

