/*
 * Hauppage HD PVR USB driver - video 4 linux 2 interface
 *
 * Copyright (C) 2008      Janne Grunau (j@jannau.net)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kref.h>
#include <linux/uaccess.h>
#include <linux/usb.h>
#include <linux/mutex.h>

#include <linux/videodev2.h>

#include "hdpvr.h"


int hdpvr_config_call(struct hdpvr_device *dev, uint value, u8 valbuf)
{
	int ret;
	char request_type = 0x38, snd_request = 0x01;
	char *buf = kmalloc(1, GFP_KERNEL);
	if (!buf)
		return -ENOMEM;

	msleep(10);

	*buf = valbuf;
	ret = usb_control_msg(dev->udev,
			      usb_sndctrlpipe(dev->udev, 0),
			      snd_request, 0x00 | request_type,
			      value, CTRL_DEFAULT_INDEX,
			      buf, 1, 10000);

	hdpvr_info("config call request returned %d", ret);

	kfree(buf);

	return ret < 0 ? ret : 0;
}

struct hdpvr_video_info *get_video_info(struct hdpvr_device *dev)
{
	u8 *buffer;
	struct hdpvr_video_info *vidinf = NULL;
#ifdef HDPVR_DEBUG
	char print_buf[15];
#endif
	int ret;

	buffer = kzalloc(5, GFP_KERNEL);
	if (!buffer) {
		err("out of memory");
		goto err;
	}
	vidinf = kzalloc(sizeof(struct hdpvr_video_info), GFP_KERNEL);
	if (!vidinf) {
		err("out of memory");
		goto err;
	}

	ret = usb_control_msg(dev->udev,
			      usb_rcvctrlpipe(dev->udev, 0),
			      0x81, 0x80 | 0x38,
			      0x1400, 0x0003,
			      buffer, 5,
			      1000);
	if (ret == 5) {
		vidinf->width	= buffer[1] << 8 | buffer[0];
		vidinf->height	= buffer[3] << 8 | buffer[2];
		vidinf->fps	= buffer[4];
	}

#ifdef HDPVR_DEBUG
	if (hdpvr_debug & MSG_INFO) {
		print_bytes(print_buf, buffer, 5);
		hdpvr_info("get video info returned: %d, %s", ret, print_buf);
	}
#endif

	if (!vidinf->width || !vidinf->height || !vidinf->fps) {
		kfree(vidinf);
		vidinf = NULL;
	}
err:
	kfree(buffer);
	return vidinf;
}

int get_input_lines_info(struct hdpvr_device *dev)
{
	u8 buffer[3];
#ifdef HDPVR_DEBUG
	char print_buf[9];
#endif
	int ret;

	ret = usb_control_msg(dev->udev,
			      usb_rcvctrlpipe(dev->udev, 0),
			      0x81, 0x80 | 0x38,
			      0x1800, 0x0003,
			      buffer, 3,
			      1000);

#ifdef HDPVR_DEBUG
	if (hdpvr_debug & MSG_INFO) {
		print_bytes(print_buf, buffer, 3);
		hdpvr_info("get input lines info returned: %d, %s", ret,
			   print_buf);
	}
#endif
	return buffer[1] << 8 | buffer[0];
}


int hdpvr_set_bitrate(struct hdpvr_device *dev)
{
	int ret;
	char *buf;

	buf = kzalloc(4, GFP_KERNEL);
	if (!buf) {
		err("out of memory");
		ret = -ENOMEM;
		goto err;
	}

	buf[0] = dev->options.bitrate;
	buf[2] = dev->options.peak_bitrate;

	ret = usb_control_msg(dev->udev,
			      usb_sndctrlpipe(dev->udev, 0),
			      0x01, 0x38, CTRL_BITRATE_VALUE,
			      CTRL_DEFAULT_INDEX, buf, 4, 1000);

err:
	kfree(buf);
	return ret;
}

int hdpvr_set_audio(struct hdpvr_device *dev, u8 input,
		    enum v4l2_mpeg_audio_encoding codec)
{

	int ret = 0, buflength = 2;
	char *buf = NULL;

	if (dev->flags & HDPVR_FLAG_AC3_CAP) {
		buf = kzalloc(buflength, GFP_KERNEL);
		if (!buf) {
			hdpvr_err("out of memory");
			ret = -ENOMEM;
			goto error;
		}

		buf[0] = input;
		if (codec == V4L2_MPEG_AUDIO_ENCODING_AAC)
			buf[1] = 0;
		else if (codec == V4L2_MPEG_AUDIO_ENCODING_AC3)
			buf[1] = 1;
		else {
			hdpvr_err("invalid audio codec %d", codec);
			ret = -EINVAL;
			goto error;
		}

		ret = usb_control_msg(dev->udev,
				      usb_sndctrlpipe(dev->udev, 0),
				      0x01, 0x38, CTRL_AUDIO_INPUT_VALUE,
				      CTRL_DEFAULT_INDEX, buf, buflength, 1000);
		if (ret == buflength)
			ret = 0;
	} else
		ret = hdpvr_config_call(dev, CTRL_AUDIO_INPUT_VALUE,
					dev->options.audio_input+1);
error:
	kfree(buf);
	return ret;
}

int hdpvr_set_options(struct hdpvr_device *dev)
{
       hdpvr_config_call(dev, CTRL_VIDEO_STD_TYPE, dev->options.video_std);

       hdpvr_config_call(dev, CTRL_VIDEO_INPUT_VALUE,
			 dev->options.video_input+1);

       hdpvr_set_audio(dev, dev->options.audio_input+1,
		       dev->options.audio_codec);

       hdpvr_set_bitrate(dev);
       hdpvr_config_call(dev, CTRL_BITRATE_MODE_VALUE,
			 dev->options.bitrate_mode);
       hdpvr_config_call(dev, CTRL_GOP_MODE_VALUE, dev->options.gop_mode);

       hdpvr_config_call(dev, CTRL_BRIGHTNESS, dev->options.brightness);
       hdpvr_config_call(dev, CTRL_CONTRAST,   dev->options.contrast);
       hdpvr_config_call(dev, CTRL_HUE,        dev->options.hue);
       hdpvr_config_call(dev, CTRL_SATURATION, dev->options.saturation);
       hdpvr_config_call(dev, CTRL_SHARPNESS,  dev->options.sharpness);

       return 0;
}
